import collections
import json
import pathlib

sols = {}
root = pathlib.Path("solutions")

bylvl = collections.defaultdict(list)

for sol in json.load((root / "solutions.json").open()):
    if "type" not in sol and sol["legal"] and sol["worky"]:
        bylvl[sol["levelNumber"]].append(sol)

missing = set(lvl["number"]
              for lvl in json.load(open("hrm/levels.json"))
              if not lvl.get("cutscene", False)) - set(bylvl)
assert not missing, missing


def solrank(sol):
    return sol["size"], sol["steps"], -sol["successRatio"]


for lvl, group in bylvl.items():
    if any(sol["author"] == "atesgoral" for sol in group):
        bylvl[lvl] = group = [sol for sol in group
                              if sol["author"] == "atesgoral"]
    group.sort(key=solrank)
    sols[lvl] = group[0] | {"source": (root / group[0]["path"]).read_text()}

with open("hrm/solutions.json", "w") as out:
    json.dump(sols, out, indent=2)



import inspect
import json
import pathlib
import time

from typing import Union
from rich import print as rprint
from rich.status import Status
from rich.text import Text

from .parse import parse as hrmparse

colors = {"inbox": "green",
          "outbox": "green",
          "copyfrom": "red",
          "copyto": "red",
          "add": "yellow",
          "sub": "yellow",
          "bumpup": "yellow",
          "bumpdn": "yellow",
          "jump": "blue",
          "jumpz": "blue",
          "jumpn": "blue"}


class Logger:
    def __init__(self, status, width):
        self.s = status
        self.c = status.console
        self.w = width

    def __enter__(self):
        return self

    def __exit__(self, *_):
        pass

    _log_hands = {"inbox", "copyfrom", "copyto", "add", "sub",
                  "bumpup", "bumpdn"}

    def __call__(self, op, args, hrm, err=None):
        opargs = Text(" ".join(f"{x}" for x in (op, *args)),  style=colors[op])
        opargs.align("left", self.w)
        text = [opargs]
        if err is not None:
            more = Text.from_markup(f"[bold red]{err}[/]")
        elif op in self._log_hands:
            more = Text(f"{hrm.hands}")
        elif op == "outbox" and hrm.outbox:
            more = Text(f"{hrm.outbox[-1]}")
        else:
            more = None
        if more is not None:
            text.append(Text.from_markup(" [dim]=>[/] "))
            text.append(more)
        self.c.print(Text.assemble(*text))


class HRMError(Exception):
    @classmethod
    def check(cls, cond, message):
        if not cond:
            raise cls(message)


class HRM (object):
    def __init__(self, prog, labels):
        self.prog = tuple((op.lower(), *args) for op, *args in prog)
        self.labels = dict(labels)
        self.check()

    @classmethod
    def parse(cls, src):
        return cls(*hrmparse(src))

    @classmethod
    def level(cls, level):
        path = pathlib.Path(__file__).parent / "levels.json"
        for lvl in json.load(path.open()):
            if lvl["number"] == level:
                return lvl
        raise ValueError(f"level {level} not found")

    @classmethod
    def from_level(cls, level):
        strlvl = str(level)
        path = pathlib.Path(__file__).parent / "solutions.json"
        solutions = json.load(path.open())
        if strlvl not in solutions:
            raise ValueError(f"missing level {level!r}")
        sol = solutions[strlvl]
        lvl = cls.level(level)
        if lvl is None:
            raise ValueError(f"missing level {level}")
        if "floor" in lvl and "tiles" in lvl["floor"]:
            floor = lvl["floor"]["tiles"]
        else:
            floor = []
        inbox = lvl["examples"][0]["inbox"]
        return cls.parse(sol["source"]), inbox, floor

    def runlevel(self, level, example=0, verbose=False):
        if isinstance(level, int):
            level = self.level(level)
        if "floor" in level and "tiles" in level["floor"]:
            floor = level["floor"]["tiles"]
        else:
            floor = []
        return self(level["examples"][example]["inbox"], floor, verbose)

    def _dummy_log(self, *_):
        pass

    def iter(self, inbox, floor=[], log=None):
        if log is None:
            log = self._dummy_log
        self.state = {}
        self.state.update(ip=0, hands=None)
        if isinstance(floor, dict):
            self.state.update((int(k), v) for k, v in floor.items())
        else:
            self.state.update(enumerate(floor))
        self.inbox = list(inbox)
        self.outbox = []
        while 0 <= self.ip < len(self.prog):
            yield self.ip
            op, *args = self.prog[self.ip]
            self.ip += 1
            handler = getattr(self, f"op_{op}")
            try:
                if handler(*args):
                    log(op, args, self, "STOP")
                    break
            except HRMError as err:
                log(op, args, self, err)
                raise
            log(op, args, self)

    def __call__(self, inbox, floor=[], verbose=False, delay=0.0):
        if verbose:
            rprint("[bold green]INBOX:[/]",
                   *(Text(f"{i}") for i in inbox))
            width = max(len(op) + sum(len(str(a)) for a in args) + len(args)
                        for op, *args in self.prog)
            with Status("working...") as status, Logger(status, width) as log:
                for _ in self.iter(inbox, floor, log):
                    time.sleep(delay)
            rprint("[bold green]OUTBOX:[/]",
                   *(Text(f"{i}") for i in self.outbox))
        else:
            list(self.iter(inbox, floor))
        return self.outbox

    @property
    def ip(self):
        return self["ip"]

    @ip.setter
    def ip(self, value):
        self["ip"] = value

    @property
    def hands(self):
        return self["hands"]

    @hands.setter
    def hands(self, value):
        self["hands"] = value

    def __getitem__(self, addr):
        if addr == "ip":
            return self.state["ip"]
        elif addr == "hands":
            return self.state["hands"]
        elif isinstance(addr, int):
            HRMError.check(self.state.get(addr, None) is not None,
                           f"tile {addr} is empty")
            return self.state[addr]
        elif isinstance(addr, list) and len(addr) == 1 \
                and isinstance(addr[0], int):
            return self[self[addr[0]]]
        else:
            raise ValueError(f"invalid address {addr!r}")

    def __setitem__(self, addr, value):
        update = getattr(self, "update", {})
        if addr == "ip":
            self.state["ip"] = update["ip"] = value
        elif addr == "hands":
            self.state["hands"] = update["hands"] = value
        elif isinstance(addr, int):
            self.state[addr] = update[addr] = value
        elif isinstance(addr, list) and len(addr) == 1 \
                and isinstance(addr[0], int):
            a = self[addr[0]]
            self[a] = update[a] = value
        else:
            raise ValueError(f"invalid address {addr!r}")

    def check(self):
        for op, *args in self.prog:
            fun = getattr(self, f"op_{op.lower()}", None)
            HRMError.check(fun is not None, f"unknown operation {op}")
            assert fun is not None  # for type checker
            sig = inspect.signature(fun)
            try:
                bound = sig.bind(*args)
            except TypeError:
                raise HRMError(f"invalid arguments for {op}: {args}")
            for name, value in bound.arguments.items():
                annot = sig.parameters[name].annotation
                HRMError.check(isinstance(value, annot),
                               f"invalid argument for {op}: {value}")
                if annot is str:
                    HRMError.check(value in self.labels,
                                   f"undefined label {value}")

    def op_inbox(self):
        if self.inbox:
            self.hands = self.inbox.pop(0)
        else:
            return True

    def op_outbox(self):
        HRMError.check(self.hands is not None, f"you don't hold any value")
        self.outbox.append(self.hands)
        self.hands = None

    def op_copyfrom(self, addr: Union[int, list]):
        self.hands = self[addr]

    def op_copyto(self, addr: Union[int, list]):
        HRMError.check(self.hands is not None, f"you don't hold any value")
        self[addr] = self.hands

    def op_add(self, addr: int):
        HRMError.check(self.hands is not None, f"you don't hold any value")
        HRMError.check(isinstance(self.hands, int),
                       f"cannot add to value {self.hands!r}")
        val = self[addr]
        HRMError.check(isinstance(val, int), f"cannot add value {val!r}")
        self.hands += val

    def op_sub(self, addr: int):
        HRMError.check(self.hands is not None, f"you don't hold any value")
        val = self[addr]
        if isinstance(self.hands, int) and isinstance(val, int):
            self.hands -= val
        elif isinstance(self.hands, str) and isinstance(val, str):
            self.hands = ord(self.hands) - ord(val)
        else:
            raise HRMError(f"cannot sub {val!r} from {self.hands!r}")

    def op_bumpup(self, addr: int):
        val = self[addr]
        HRMError.check(isinstance(val, int),
                       f"cannot increment value {self.hands!r}")
        self.hands = self[addr] = val + 1

    def op_bumpdn(self, addr: int):
        val = self[addr]
        HRMError.check(isinstance(val, int),
                       f"cannot decrement value {self.hands!r}")
        self.hands = self[addr] = val - 1

    def op_jump(self, lbl: str):
        HRMError.check(lbl in self.labels, f"labels {lbl} is not defined")
        pos = self.labels[lbl]
        HRMError.check(0 <= pos <= len(self.prog), f"invalid program position")
        if pos == len(self.prog):
            return True
        self.ip = pos

    def op_jumpz(self, lbl: str):
        HRMError.check(lbl in self.labels, f"labels {lbl} is not defined")
        pos = self.labels[lbl]
        HRMError.check(self.hands is not None, f"you don't hold any value")
        HRMError.check(0 <= pos <= len(self.prog), f"invalid program position")
        if self.hands == 0:
            if pos == len(self.prog):
                return True
            self.ip = pos

    def op_jumpn(self, lbl: str):
        HRMError.check(lbl in self.labels, f"labels {lbl} is not defined")
        pos = self.labels[lbl]
        HRMError.check(self.hands is not None, f"you don't hold any value")
        HRMError.check(0 <= pos <= len(self.prog), f"invalid program position")
        if self.hands < 0:
            if pos >= len(self.prog):
                return True
            self.ip = pos



# coding: utf-8

import json, collections, operator, sys, pathlib
import tqdm

from colorama import Fore as F, Style as S
from IPython.core import ultratb

from hrm import HRM
from hrm.tikz import draw

log = tqdm.tqdm(sorted(json.load(open("solutions/solutions.json")),
                       key=operator.itemgetter("levelNumber")))
pdf_base = pathlib.Path("pdf")
errors = collections.defaultdict(int)

for sol in log :
    path = f"solutions/{sol['path']}"
    try :
        hrm = HRM.parse(path)
    except :
        log.write(f"{F.RED}parse error:{S.RESET_ALL} {sol['path']}")
        errors[f"{F.RED}parse error"] += 1
    # draw code
    pdf = (pdf_base / sol["path"]).with_suffix(".pdf")
    pdf.parent.mkdir(parents=True, exist_ok=True)
    try :
        draw(path, pdf)
    except :
        log.write(f"{F.YELLOW}draw failed:{S.RESET_ALL} {sol['path']}")
        errors[f"{F.YELLOW}draw failed"] += 1
    # exec code
    num = sol["levelNumber"]
    lvl = hrm.level(num)
    for ex, example in enumerate(lvl["examples"]) :
        try :
            out = hrm.runlevel(lvl, ex)
        except AssertionError as err :
            errors[f"{F.RED}invalid solution"] += 1
            log.write(f"{F.RED}{S.BRIGHT}invalid:{S.RESET_ALL} {sol['path']}")
            log.write(f"  😡 {F.RED}{S.DIM}{err}{S.RESET_ALL}")
            break
        except :
            errors[f"{F.RED}crash"] = 1
            log.write(f"{F.RED}{S.BRIGHT}crashed:{S.RESET_ALL} {sol['path']}")
            vtb = ultratb.VerboseTB(color_scheme="Linux")
            log.write("\n".join(vtb.structured_traceback(*sys.exc_info())))
            break
        if sol["successRatio"] == 1 and sol["worky"] and out != example["outbox"] :
            errors["{F.YELLOW}failed"] += 1
            log.write(f"{F.YELLOW}failed:{S.RESET_ALL} {sol['path']}")

for err, count in errors.items() :
    print(f"=> {err}:{S.RESET_ALL} {count}")



import random
import re
import string

from typing import Annotated, Optional

from typer import Typer, Option, Argument, Exit
from rich import print as rprint
from rich.text import Text

from . import HRM, HRMError
from .tui import main as tui


app = Typer(context_settings={"help_option_names": ["-h", "--help"]})


def intval(txt):
    try:
        return int(txt)
    except ValueError:
        pass


def parse_tiles(text):
    items = []
    for val in text.split(","):
        if val == "":
            items.append(None)
        elif (num := intval(val)) is not None:
            items.append(num)
        elif len(val) == 1:
            items.append(val)
        else:
            raise ValueError("invalid value {val!r}")
    return items


def parse_inbox(text):
    return [val for val in parse_tiles(text) if val is not None]


def gen_inbox(length, neg, chars, bound):
    values = []
    values.extend(range(bound+1))
    if neg:
        values.extend([-v for v in values])
    if chars:
        values.extend(string.ascii_uppercase)
    return [random.choice(values) for _ in range(length)]


def build(prog, inbox, tiles, length, negative, chars, maxval):
    if match := re.match(r"^(lvl|level):(\d+)$", prog, re.I):
        hrm, inbox, tiles = HRM.from_level(int(match.group(2)))
    else:
        hrm = HRM.parse(prog)
        if inbox is None:
            inbox = gen_inbox(length, negative, chars, maxval)
    assert isinstance(inbox, list)
    return hrm, inbox, tiles


@app.command(help="fully run a program and print its outbox")
def run(prog: Annotated[
            str,
            Argument(
                help="program to run: either a PATH to source or 'lvl:NUM'")],
        verbose: Annotated[
            bool,
            Option(
                "-v", "--verbose",
                help="print operations as they are executed"
            )
        ] = False,
        inbox: Annotated[
            Optional[list[str | int]],
            Option(
                "-i", "--inbox",
                metavar="LIST",
                parser=parse_inbox,
                help="inbox to use instead of generated one (eg, '1,2,1,B,3')"
            )] = None,
        tiles: Annotated[
            list[str | int],
            Option(
                "-t", "--tiles",
                metavar="LIST",
                parser=parse_tiles,
                help="tiles to be used (eg, '1,,,0' with empty slots allowed)"
            )] = [],
        delay: Annotated[
            float,
            Option(
                "-d", "--delay",
                metavar="DELAY",
                help="wait DELAY seconds after each operation"
            )] = 0.0,
        negative: Annotated[
            bool,
            Option(
                "-n", "--negative",
                help="allow negative numbers in generated inbox"
            )] = False,
        chars: Annotated[
            bool,
            Option(
                "-c", "--chars",
                help="allow characters in generated inbox"
            )] = False,
        length: Annotated[
            int,
            Option(
                "-l", "--length",
                metavar="INT",
                help="generate inbox with INT items"
            )] = 10,
        maxval: Annotated[
            int,
            Option(
                "-m", "--max",
                metavar="INT",
                help="generate inbox with |values| <= INT"
            )] = 10):
    hrm, inbox, tiles = build(prog, inbox, tiles,
                              length, negative, chars, maxval)
    try:
        outbox = hrm(inbox, tiles, verbose, delay)
        if not verbose:
            print(*outbox)
    except HRMError:
        raise Exit(1)


@app.command(help="run a program interactively")
def play(prog: Annotated[
            str,
            Argument(
                help="program to run: either a PATH to source or 'lvl:NUM'")],
         inbox: Annotated[
             Optional[list[str | int]],
             Option(
                 "-i", "--inbox",
                 metavar="LIST",
                 parser=parse_inbox,
                 help="inbox to use instead of generated one (eg, '1,2,1,B,3')"
             )] = None,
         tiles: Annotated[
             list[str | int],
             Option(
                 "-t", "--tiles",
                 metavar="LIST",
                 parser=parse_tiles,
                 help="tiles to be used (eg, '1,,,0' with empty slots allowed)"
             )] = [],
         delay: Annotated[
             float,
             Option(
                 "-d", "--delay",
                 metavar="DELAY",
                 help="wait DELAY seconds after each operation"
             )] = 0.0,
         negative: Annotated[
             bool,
             Option(
                 "-n", "--negative",
                 help="allow negative numbers in generated inbox"
             )] = False,
         chars: Annotated[
             bool,
             Option(
                 "-c", "--chars",
                 help="allow characters in generated inbox"
             )] = False,
         length: Annotated[
             int,
             Option(
                 "-l", "--length",
                 metavar="INT",
                 help="generate inbox with INT items"
             )] = 10,
         maxval: Annotated[
             int,
             Option(
                 "-m", "--max",
                 metavar="INT",
                 help="generate inbox with |values| <= INT"
             )] = 10):
    hrm, inbox, tiles = build(prog, inbox, tiles,
                              length, negative, chars, maxval)
    tui(hrm, inbox, tiles)


if __name__ == "__main__":
    app()
